#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 16
void NES_mapper16::Reset()
{
  patch = 0;

  // for Famicom Jump 2
  if(sum_16k_rom == 0x1d9f9c)
  {
    patch = 1;
  }

  if(!patch)
  {
    NES_6502::Context context;
    parent_NES->cpu->GetContext(&context);
    context.mem_page[3] = br_out;
    parent_NES->cpu->SetContext(&context);
    memset(br_out, 0x00, 0x2000);
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  regs[0] = 0;
  regs[1] = 0;
  regs[2] = 0;

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper16::MemoryWrite(uint32 addr, uint8 data)
{
  if(patch)
  {
    MemoryWrite3(addr, data);
  }
  else
  {
    MemoryWrite2(addr, data);
  }
}

void NES_mapper16::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(patch)
  {
    //MemoryWrite3(addr, data);
  }
  else
  {
    MemoryWrite2(addr, data);
  }
}

void NES_mapper16::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter <= 114)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = 0;
      irq_enabled = 0;
    }
    else
    {
      irq_counter -= 114;
    }
  }
}

void NES_mapper16::MemoryWrite2(uint32 addr, uint8 data)
{
  switch(addr & 0x000F)
  {
    case 0x0000:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0x0001:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0x0002:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0x0003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0x0004:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0x0005:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0x0006:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0x0007:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0x0008:
      {
        set_CPU_bank4(data*2+0);
        set_CPU_bank5(data*2+1);
      }
      break;

    case 0x0009:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 2)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0x000A:
      {
        irq_enabled = data & 0x01;
        irq_counter = irq_latch;
      }
      break;

    case 0x000B:
      {
        irq_latch = (irq_latch & 0xFF00) | data;
      }
      break;

    case 0x000C:
      {
         irq_latch = ((uint32)data << 8) | (irq_latch & 0x00FF);
      }
      break;

    case 0x000D:
      {
        if(data == 0x80)
        {
          // reset
          br_addr = 0x00;
          br_mode = 0x00;
          br_wbit = 0x00;
          br_rbit = 0x00;
        }
        else if(br_cmd[3] == 0x00 && br_cmd[2] == 0x20 && br_cmd[1] == 0x60 &&
                br_cmd[0] == 0x40 && data == 0x00)
        {
          // reset ?
        }
        else if(br_cmd[3] == 0x00 && br_cmd[2] == 0x40 && br_cmd[1] == 0x60 &&
                br_cmd[0] == 0x20 && data == 0x00)
        {
          // start
          br_wbit = 0x01;
          br_rbit = 0x01;
          br_data = 0;
          br_mode = 0;
        }
        else if (br_cmd[0] == 0x60 && data == 0xE0)
        {
          if(!br_mode)
          {
            // sync
            br_wbit = 0x01;
            br_rbit = 0x01;
            br_data = 0;
            br_mode = 1;
            memset(br_out, 0x00, 0x2000);
          }
          else
          {
            // read
            br_data = parent_NES->ReadSaveRAM(br_addr);
            if(br_data & br_rbit)
            {
              memset(br_out, 0xFF, 0x2000);
            }
            else
            {
              memset(br_out, 0x00, 0x2000);
            }
            br_rbit <<= 1;
            br_wbit = 0x00;
          }
        }
        else if(br_cmd[1] == 0x00 && br_cmd[0] == 0x20 && data == 0x00)
        {
          // write 0
          br_data &= 0xFF - br_wbit;
          if(br_wbit == 0x80)
          {
            if(br_mode)
            {
              parent_NES->WriteSaveRAM(br_addr, br_data);
            }
            else
            {
              br_addr = br_data & 0x7F;
            }
            br_wbit = 0x00;
          }
          else
          {
            br_wbit <<= 1;
          }
          br_rbit = 0x00;
        }
        else if(br_cmd[3] == 0x00 && br_cmd[2] == 0x40 && br_cmd[1] == 0x60 &&
                br_cmd[0] == 0x40 && data == 0x00)
        {
          // write 1
          br_data |= br_wbit;
          if(br_wbit == 0x80)
          {
            if(br_mode)
            {
              parent_NES->WriteSaveRAM(br_addr, br_data);
            }
            else
            {
              br_addr = br_data & 0x7F;
            }
            br_wbit = 0x00;
          }
          else
          {
            br_wbit <<= 1;
          }
          br_rbit = 0x00;
        }
        br_cmd[3] = br_cmd[2];
        br_cmd[2] = br_cmd[1];
        br_cmd[1] = br_cmd[0];
        br_cmd[0] = data;
      }
      break;
  }
}

void NES_mapper16::MemoryWrite3(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
    case 0x8001:
    case 0x8002:
    case 0x8003:
      {
        regs[0] = data & 0x01;
        set_CPU_bank4(regs[0]*0x20+regs[2]*2+0);
        set_CPU_bank5(regs[0]*0x20+regs[2]*2+1);
      }
      break;

    case 0x8004:
    case 0x8005:
    case 0x8006:
    case 0x8007:
      {
        regs[1] = data & 0x01;
        set_CPU_bank6(regs[1]*0x20+0x1E);
        set_CPU_bank7(regs[1]*0x20+0x1F);
      }
      break;

    case 0x8008:
      {
        regs[2] = data;
        set_CPU_bank4(regs[0]*0x20+regs[2]*2+0);
        set_CPU_bank5(regs[0]*0x20+regs[2]*2+1);
        set_CPU_bank6(regs[1]*0x20+0x1E);
        set_CPU_bank7(regs[1]*0x20+0x1F);
      }
      break;

    case 0x8009:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 2)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0x800A:
      {
        irq_enabled = data & 0x01;
        irq_counter = irq_latch;
      }
      break;

    case 0x800B:
      {
        irq_latch = (irq_latch & 0xFF00) | data;
      }
      break;

    case 0x800C:
      {
         irq_latch = ((uint32)data << 8) | (irq_latch & 0x00FF);
      }
      break;

    case 0x800D:
      {
        //write protect
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 66
void NES_mapper66::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
  set_PPU_banks(0,1,2,3,4,5,6,7);

  // ignore frame-IRQ for Paris-Dakar Ralley Special
  if(sum_16k_rom == 0x1de79b)
  {
    parent_NES->frame_irq_disenabled = 1;
  }
}

void NES_mapper66::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  uint8 chr_bank = data & 0x0F;
  uint8 prg_bank = (data & 0xF0) >> 4;

  set_CPU_bank4(prg_bank*4+0);
  set_CPU_bank5(prg_bank*4+1);
  set_CPU_bank6(prg_bank*4+2);
  set_CPU_bank7(prg_bank*4+3);

  set_PPU_bank0(chr_bank*8+0);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);
}

void NES_mapper66::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 chr_bank = data & 0x0F;
  uint8 prg_bank = (data & 0xF0) >> 4;

  set_CPU_bank4(prg_bank*4+0);
  set_CPU_bank5(prg_bank*4+1);
  set_CPU_bank6(prg_bank*4+2);
  set_CPU_bank7(prg_bank*4+3);

  set_PPU_bank0(chr_bank*8+0);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 70
void NES_mapper70::Reset()
{
  patch = 0;

  // for Kamen Rider Club, Family Trainer - Manhattan Police
  if(sum_16k_rom == 0x16ec09 || sum_16k_rom == 0x18bcfa)
  {
    patch = 1;
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
}

void NES_mapper70::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 chr_bank = data & 0x0F;
  uint8 prg_bank = (data & 0x70) >> 4;

  set_CPU_bank4(prg_bank*2);
  set_CPU_bank5(prg_bank*2+1);

  set_PPU_bank0(chr_bank*8);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);

  if(patch)
  {
    if(data & 0x80)
    {
      set_mirroring(NES_PPU::MIRROR_HORIZ);
    }
    else
    {
      set_mirroring(NES_PPU::MIRROR_VERT);
    }
  }
  else
  {
    if(data & 0x80)
    {
      set_mirroring(1,1,1,1);
    }
    else
    {
      set_mirroring(0,0,0,0);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 96
void NES_mapper96::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  vbank0 = vbank1 = 0;
  sync_PPU_banks();

  set_mirroring(0,0,0,0);
}

void NES_mapper96::PPU_Latch_Address(uint32 addr)
{
  if((addr & 0xF000) == 0x2000)
  {
    vbank1 = (addr & 0x0300) >> 8;
    sync_PPU_banks();
  }
}

void NES_mapper96::MemoryWrite(uint32 addr, uint8 data)
{
  set_CPU_bank4((data & 0x03) * 4 + 0);
  set_CPU_bank5((data & 0x03) * 4 + 1);
  set_CPU_bank6((data & 0x03) * 4 + 2);
  set_CPU_bank7((data & 0x03) * 4 + 3);

  vbank0 = (data & 0x04) >> 2;
  sync_PPU_banks();
}

void NES_mapper96::sync_PPU_banks()
{
  parent_NES->ppu->PPU_VRAM_banks[0] = vram + vbank0 * 0x4000 + vbank1 * 0x1000 + 0x000;
  parent_NES->ppu->PPU_VRAM_banks[1] = vram + vbank0 * 0x4000 + vbank1 * 0x1000 + 0x400;
  parent_NES->ppu->PPU_VRAM_banks[2] = vram + vbank0 * 0x4000 + vbank1 * 0x1000 + 0x800;
  parent_NES->ppu->PPU_VRAM_banks[3] = vram + vbank0 * 0x4000 + vbank1 * 0x1000 + 0xC00;
  parent_NES->ppu->PPU_VRAM_banks[4] = vram + vbank0 * 0x4000 + 0x3000;
  parent_NES->ppu->PPU_VRAM_banks[5] = vram + vbank0 * 0x4000 + 0x3400;
  parent_NES->ppu->PPU_VRAM_banks[6] = vram + vbank0 * 0x4000 + 0x3800;
  parent_NES->ppu->PPU_VRAM_banks[7] = vram + vbank0 * 0x4000 + 0x3C00;
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 188
void NES_mapper188::Reset()
{
  // set CPU bank pointers
  if(num_8k_ROM_banks > 16)
  {
    set_CPU_banks(0,1,14,15);
  }
  else
  {
    set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  }

  NES_6502::Context context;
  parent_NES->cpu->GetContext(&context);
  context.mem_page[3] = dummy;
  parent_NES->cpu->SetContext(&context);

  dummy[0] = 0x03; // ?
}

// Coded by rinao
void NES_mapper188::MemoryWrite(uint32 addr, uint8 data)
{
  if (data)
  {
    if (data & 0x10)
    {
      data &= 0x07;
      set_CPU_bank4(data*2);
      set_CPU_bank5(data*2+1);
    }
    else
    {
      set_CPU_bank4(data*2+16);
      set_CPU_bank5(data*2+17);
    }
  }
  else
  {
    if (num_8k_ROM_banks == 0x10)
    {
      set_CPU_bank4(14);
      set_CPU_bank5(15);
    }
    else
    {
      set_CPU_bank4(16);
      set_CPU_bank5(17);
    }
  }
  //data &= num_8k_ROM_banks-1;
  //set_CPU_banks(data*2,(data*2)+1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
}
/////////////////////////////////////////////////////////////////////

#endif
